@@ -1886,6 +1886,15 @@ unmangle_sender_field (ExchangeData *ed, E2kOperation *op, const gchar *uri, gch
 	return E2K_HTTP_OK;
 }
 
+static void
+foreign_new_folder_cb (ExchangeAccount *account, EFolder *folder, GPtrArray *folders)
+{
+	g_return_if_fail (folder != NULL);
+	g_return_if_fail (folders != NULL);
+
+	g_ptr_array_add (folders, folder);
+}
+
 static void
 get_folder_info_data (ExchangeData *ed, const gchar *top, guint32 store_flags, GPtrArray **names, GPtrArray **uris, GArray **unread, GArray **flags)
 {
@@ -1948,6 +1957,11 @@ get_folder_info_data (ExchangeData *ed, const gchar *top, guint32 store_flags, G
 	}
 
 	if (folders) {
+		guint new_folder_handler_id = 0;
+
+		if (ed->new_folder_id == 0)
+			new_folder_handler_id = g_signal_connect (ed->account, "new_folder", G_CALLBACK (foreign_new_folder_cb), folders);
+
 		for (i = 0; i < folders->len; i++) {
 			folder = folders->pdata[i];
 			hier = e_folder_exchange_get_hierarchy (folder);
@@ -1990,7 +2004,7 @@ get_folder_info_data (ExchangeData *ed, const gchar *top, guint32 store_flags, G
 					}
 					break;
 				case EXCHANGE_HIERARCHY_FOREIGN:
-					if (folder_flags & CAMEL_FOLDER_NOSELECT && ed->new_folder_id == 0) {
+					if ((folder_flags & CAMEL_FOLDER_NOSELECT) != 0 && ed->new_folder_id == 0) {
 						/* Rescan the hierarchy - as we don't rescan
 						   foreign hierarchies anywhere for mailer and
 						   only when we are starting up
@@ -2031,6 +2045,8 @@ get_folder_info_data (ExchangeData *ed, const gchar *top, guint32 store_flags, G
 			g_array_append_val (*flags, folder_flags);
 		}
 
+		if (new_folder_handler_id)
+			g_signal_handler_disconnect (ed->account, new_folder_handler_id);
 		g_ptr_array_free (folders, TRUE);
 	}
 }
@@ -2079,17 +2095,17 @@ camel_exchange_utils_connect (CamelService *service,
 	E2kContext *ctx;
 	guint32 retval = 1;
 	const gchar *uri;
-	gint mode;
+	struct update_linestatus ul;
 
 	g_return_val_if_fail (ed != NULL, FALSE);
 	g_return_val_if_fail (status != NULL, FALSE);
 
-	mode = is_online (ed);
+	ul.linestatus = is_online (ed);
 
 	account = ed->account;
-	if (mode == ONLINE_MODE)
+	if (ul.linestatus == ONLINE_MODE)
 		exchange_account_set_online (account);
-	else if (mode == OFFLINE_MODE)
+	else if (ul.linestatus == OFFLINE_MODE)
 		exchange_account_set_offline (account);
 
 	ctx = exchange_account_get_context (account);
@@ -2097,10 +2113,10 @@ camel_exchange_utils_connect (CamelService *service,
 		ctx = exchange_account_connect (account, pwd, &result);
 	}
 
-	if (!ctx && mode == ONLINE_MODE) {
+	if (!ctx && ul.linestatus == ONLINE_MODE) {
 		retval = 0;
 		goto end;
-	} else if (mode == OFFLINE_MODE) {
+	} else if (ul.linestatus == OFFLINE_MODE) {
 		goto end;
 	}
 
@@ -2117,9 +2133,11 @@ camel_exchange_utils_connect (CamelService *service,
 	/* Will be used for offline->online transition to initialize things for
 	   the first time */
 
+	ul.estore = ed->estore;
+	ul.ex = ex;
 	g_hash_table_foreach (ed->folders_by_name,
 			      (GHFunc) folder_update_linestatus,
-			      GINT_TO_POINTER (mode));
+			      &ul);
  end:
 	*status = retval;
 
@@ -2924,6 +2942,11 @@ camel_exchange_utils_get_folder_info (CamelService *service,
 	g_return_val_if_fail (unread_counts != NULL, FALSE);
 	g_return_val_if_fail (folder_flags != NULL, FALSE);
 
+	/* use lock here to have done scanning of foreign hierarchy
+	   only once, and to not call get_folder_info_data simultaneously
+	   from more than one thread */
+	g_static_rec_mutex_lock (&ed->changed_msgs_mutex);
+
 	get_folder_info_data (ed, top, store_flags, folder_names, folder_uris, unread_counts, folder_flags);
 
 	if (ed->new_folder_id == 0) {
@@ -2931,6 +2954,8 @@ camel_exchange_utils_get_folder_info (CamelService *service,
 		ed->removed_folder_id = g_signal_connect (ed->account, "removed_folder", G_CALLBACK (account_removed_folder), ed);
 	}
 
+	g_static_rec_mutex_unlock (&ed->changed_msgs_mutex);
+
 	return TRUE;
 }
 
